-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt
-----
-- by @MegaKnight
--
-- DO NOT REUSE CODE WITHOUT PERMISSION
-----

local M = {}


----
-- variables 
----
local sunsky = nil
local clouds = nil
local cloudColor = nil
local fog = nil
local rain = nil
local rainSound = nil
local forestWind = nil
local physicsWind = Point3F(0,0,0)
local isWindBursting = false --> is wind burst currently in action ?
local forestToPhysicsFactor = 25

-- is level unsupported / main menu ?
local unsupported = false

-- display messages ?
local doMessageDisplay = true


-- possible weather events with maximum values
local events = {}
	events.sunny = {} -- sunny
		events.sunny.cloudCover = 0.25
		events.sunny.fogDensity = 0.0001 --> constant value
		events.sunny.rainAmount = 0
		events.sunny.windSpeed = 0.3
	events.cloudy = {} -- cloudy
		events.cloudy.cloudCover = 0.6
		events.cloudy.fogDensity = 0.0008 --> constant value
		events.cloudy.rainAmount = 100
		events.cloudy.windSpeed = 0.45
	events.foggy = {} -- foggy
		events.foggy.cloudCover = 1.0
		events.foggy.fogDensity = 0.1
		events.foggy.rainAmount = 100
		events.foggy.windSpeed = 0.15
	events.foggyRain = {} -- foggy with rain
		events.foggyRain.cloudCover = 1.0
		events.foggyRain.fogDensity = 0.1
		events.foggyRain.rainAmount = 1500
		events.foggyRain.windSpeed = 0.2
	events.lightRain = {} 
		events.lightRain.cloudCover = 0.6
		events.lightRain.fogDensity = 0.005
		events.lightRain.rainAmount = 800
		events.lightRain.windSpeed = 0.45
	events.rainy = {} -- rainy
		events.rainy.cloudCover = 1.0
		events.rainy.fogDensity = 0.01
		events.rainy.rainAmount = 2000
		events.rainy.windSpeed = 0.5
	events.stormy = {} -- stormy
		events.stormy.cloudCover = 1.5
		events.stormy.fogDensity = 0.05
		events.stormy.rainAmount = 4000
		events.stormy.windSpeed = 1.0
		
-- current weather event --> standard value
local currentEvent = "sunny"
--local queuedEvent = "sunny" --> not used

-- reach these values when changing to event:
local cloudCoverGoal = {}
	cloudCoverGoal.sunny = 0.2
	cloudCoverGoal.cloudy = 0.5
	cloudCoverGoal.lightRain = 0.6
	cloudCoverGoal.rainy = 0.8
	cloudCoverGoal.foggy = 0.8
	cloudCoverGoal.foggyRain = 0.8
	cloudCoverGoal.stormy = 1.2
local fogGoal = {}
	fogGoal.sunny = 0.0001 --> constant value
	fogGoal.cloudy = 0.0008 --> constant value
	fogGoal.lightRain = 0.002
	fogGoal.rainy = 0.005
	fogGoal.foggy = 0.015
	fogGoal.foggyRain = 0.015
	fogGoal.stormy = 0.005
local windGoal = {}
	windGoal.sunny = 0.2
	windGoal.cloudy = 0.3
	windGoal.lightRain = 0.35
	windGoal.rainy = 0.4
	windGoal.foggy = 0.1
	windGoal.foggyRain = 0.15
	windGoal.stormy = 0.7 --> not necessesarily windy storm, but heavy rain ?
local rainGoal = {}
	rainGoal.sunny = 0
	rainGoal.cloudy = 0
	rainGoal.lightRain = 500
	rainGoal.rainy = 1000
	rainGoal.foggy = 0
	rainGoal.foggyRain = 800
	rainGoal.stormy = 2500
-- rainVolumeFactor : 2000 => 1 , 0 => 0 : rainVolumeFactor = rain.numDrops/2000

-- is event that allows random fog/rain changes ?
local isFogEvent = false
local isRainEvent = false




-----
-- ground model logic
-----

-- is rain above/below treshold?
local rainAboveTreshold = false
local rainBelowTreshold = true

-- rain treshold for ground model change
local rainTreshold = 300 --> 300 numDrops

-- find terrain paints to be affected by ground model change --> maybe, if needed
--local asphaltPaints = {} --> storage table
--local function findAsphaltPaints()
--end

-- change ground model of asphalt/concrete terrain paint to ~ASPHALT_WET/ASPHALT
-- code by @Comodor  
local function changeGroundModel(isWet)

	-- table to store values
	local groundmodel = ground_model()
	
	if isWet then -- wet asphalt
		groundmodel.staticFrictionCoefficient = 0.92
		groundmodel.slidingFrictionCoefficient = 0.55
		groundmodel.hydrodynamicFriction = 0
		groundmodel.stribeckVelocity = 5
		groundmodel.strength = 1
		groundmodel.roughnessCoefficient = 0.2
		groundmodel.fluidDensity = 0
		groundmodel.flowConsistencyIndex = 0
		groundmodel.flowBehaviorIndex = 0
		groundmodel.dragAnisotropy = 0
		groundmodel.shearStrength = 0
		groundmodel.defaultDepth = 0
		groundmodel.collisiontype = 11 --> wet asphalt
		groundmodel.skidMarks= true
		log("I", "", "Changing ground model of asphalt to wet asphalt")
	else -- normal asphalt
		groundmodel.staticFrictionCoefficient = 0.98
		groundmodel.slidingFrictionCoefficient = 0.70
		groundmodel.hydrodynamicFriction = 0
		groundmodel.stribeckVelocity = 4.5
		groundmodel.strength = 1
		groundmodel.roughnessCoefficient = 0
		groundmodel.fluidDensity = 0
		groundmodel.flowConsistencyIndex = 0
		groundmodel.flowBehaviorIndex = 0
		groundmodel.dragAnisotropy = 0
		groundmodel.shearStrength = 0
		groundmodel.defaultDepth = 0
		groundmodel.collisiontype = 10 --> asphalt
		groundmodel.skidMarks= false
		log("I", "", "Changing ground model of asphalt to asphalt")
	end
	
	-- change ground model of terrain paints
	be:setGroundModel("groundmodel_asphalt1", groundmodel)
	be:setGroundModel("asphalt", groundmodel)
	be:setGroundModel("asphalt2", groundmodel)
	be:setGroundModel("groundmodel_asphalt_old", groundmodel)
	be:setGroundModel("Asphalt", groundmodel)
	be:setGroundModel("asphalt3", groundmodel)
	be:setGroundModel("Concrete", groundmodel)
	be:setGroundModel("concrete", groundmodel)
	be:setGroundModel("concrete2", groundmodel)
end




-----
-- loading, init and unloading functions
-----

-- initialize cloud settings -> fluffy
local function initClouds()
	clouds.coverage = 0.2 
	clouds.baseColor = Point4F(1,1,1,1) 
	cloudColor = 1
	clouds.windSpeed = 0.3
	clouds.exposure = 1.5
	clouds.texture = "levels/utah/art/skies/clouds_fluffy.dds"
	clouds.height = 8.5
	clouds.isRenderEnabled = true
	--clouds:postApply() doesn't work
	
	scenetree.findObjectById(clouds:getId()):postApply()
end

-- initialize sunsky settings -> summer day
local function initSunSky()
	sunsky.skyBrightness = 600.0
	sunsky.sunSize = 0.1
	sunsky.sunScale = Point4F(1,0.827,0.725,1) -- 255,211,185,255
	sunsky.rayleighScattering = 0.0001
	sunsky.flareType = scenetree.findObject("BNG_Sunflare_3")
	sunsky.flareScale = 5.0
	sunsky.colorize = Point4F(0.243,0.392,0.678,1) -- 62,100,173,255
	sunsky.groundAlbedo = Point4F(0.89,0.666,0.408,1) -- 227,170,104,255
	sunsky.fogScale = Point4F(0.514,0.733,1,1) -- 131,187,255,255
	sunsky.enableFogFallback = false
	sunsky.exposure = 2.0
	sunsky.brightness = 1.0
	
	--scenetree.findObjectById(sunsky:getId()):postApply() --> do later
end

-- initialize fog settings -> far visibility
local function initFog()
	fog.fogColor = sunsky.fogScale
	fog.fogDensity = 0.0001 --> fog visibility
	fog.visibleDistance = 5000 --> render visibility, change together with fog density
	fog.fogAtmosphereHeight = 2000
	
	scenetree.findObjectById(fog:getId()):postApply()
	scenetree.findObjectById(sunsky:getId()):postApply() --> overwrite theLevelInfo fogScale, as it's incorrect
end

-- initialize rain settings --> TODO: snow?
local function initRain()
	rain = createObject("Precipitation")
	rain.dataBlock = scenetree.findObject("rain_medium")
	rain.numDrops = 0
	rain.splashSize = 0
	rain.splashMS = 0
	rain.animateSplashes = 0
	rain.boxWidth = 16.0
	rain.boxHeight = 16.0
	rain.boxoceanHeight = 10.0
	rain.dropSize = 1.0
	rain.doCollision = true
	rain.hitVehicles = true
	rain.rotateWithCamVel = false
	rain.followCam = true
	rain.useWind = true
	rain.minSpeed = 0.3
	rain.maxSpeed = 0.6
	rain.minMass = 4
	rain.maxMass = 5
	rain:registerObject('rain_coverage_MKDW')
	
	--scenetree.findObjectById(rain:getId()):postApply()
end

-- initialize rain sound settings
local function initRainSound()
	rainSound = createObject("SFXEmitter")
    rainSound.scale = Point3F(100, 100, 100)
    rainSound.fileName = String('/art/sound/environment/amb_rain_medium.ogg')
    rainSound.playOnAdd = true
    rainSound.isLooping = true
    rainSound.volume = 0.0
    rainSound.isStreaming = true
    rainSound.is3D = false
    rainSound:registerObject('rain_sound_MKDW')
	
	--scenetree.findObjectById(rainSound:getId()):postApply()
end

-- initialize forest wind settings --> affecting forest items and groundcover
local function initForestWind()
	forestWind.windEnabled = true
	forestWind.strength = 0.2
	forestWind.gustFrequency = 0.0
	
	scenetree.findObjectById(forestWind:getId()):postApply()
end

-- initialize wind settings --> affecting physics objects
local function initPhysicsWind()
	physicsWind = Point3F(forestWind.strength * forestToPhysicsFactor, forestWind.strength * forestToPhysicsFactor, 0)
	-- be:queueAllObjectLua('obj:setWind(1, 1, 0)') --> strength(x, y, z), probably in kph
end


------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------

-- executed when lua extension loaded -> keyboard input
local function onExtensionLoaded()
	
	-- TODO: start with random event!
	
	-- if not in level
	if not scenetree then extensions.unload('dynamicWeatherMK_weather') return end
	
	
	-- get cloud objects
	local cloudObjs = scenetree.findSubClassObjects('CloudLayer')
	if cloudObjs[1] == nil then unsupported = true extensions.unload('dynamicWeatherMK_weather') return end
	
	-- set cloud variable to first cloud in table, hide all other cloud objects
	for i, v in ipairs(cloudObjs) do
		if i == 1 then
			scenetree.findObject(cloudObjs[i]).hidden = false
			clouds = scenetree.findObject(cloudObjs[i])
		else
			scenetree.findObject(cloudObjs[i]).hidden = true
		end
	end
	
	-- initialize cloud settings
	initClouds()
	
	
	-- get sunsky
	local skyObj = scenetree.findSubClassObjects('ScatterSky')
	if skyObj[1] == nil then 
		unsupported = true extensions.unload('dynamicWeatherMK_weather') return 
	else
		sunsky = scenetree.findObject(skyObj[1]) 
		if not sunsky then unsupported = true extensions.unload('dynamicWeatherMK_weather') return end
		for i, v in ipairs(skyObj) do
			if scenetree.findObject(skyObj[i]):getId() == sunsky:getId() then scenetree.findObject(skyObj[i]).hidden = false 
			else scenetree.findObject(skyObj[i]).hidden = true
			end
		end
	end
	
	-- initialize sunsky settings
	initSunSky()
	
	
	-- get levelInfo for fog settings
	fog = scenetree.theLevelInfo
	if not fog then unsupported = true extensions.unload('dynamicWeatherMK_weather') return end
	
	-- initialize fog settings
	initFog()
	
	
	-- TODO: check for snow?
	-- hide existing rain
	local rainObjs = scenetree.findSubClassObjects("Precipitation")
	--dump(rainObjs)
	for i, v in ipairs(rainObjs) do
		-- delete if from previous activation
		if rainObjs[i] == "rain_coverage_MKDW" then
			scenetree.findObject(rainObjs[i]):delete()
		else
			scenetree.findObject(rainObjs[i]).hidden = true
		end
	end
		
	-- create and initialize rain object
	initRain()
		
	
	-- mute rain sound, if existing
	local rainSoundObjs = scenetree.findSubClassObjects("SFXEmitter")
	
	for i, v in ipairs(rainSoundObjs) do
		-- delete if from previous activation
		--dump(rainSoundObjs)
		if rainSoundObjs[i] == "rain_sound_MKDW" then
			scenetree.findObject(rainSoundObjs[i]):delete()
		elseif rainSoundObjs[i] == "rain_sound" then
			scenetree.findObject(rainSoundObjs[i]).hidden = true
		end
	end
	
	-- create and initialize rain soud emitter --> if not snow
	initRainSound()
	
	
	-- get ForestWindEmitter
	local windObjs = scenetree.findSubClassObjects("ForestWindEmitter")
	for i, v in ipairs(windObjs) do
		if i == 1 then
			forestWind = scenetree.findObject(windObjs[i])
		else
			scenetree.findObject(windObjs[i]).hidden = true
		end
	end
	
	-- create ForestWindEmitter if not existing
	if not forestWind then 
		forestWind = createObject("ForestWindEmitter")
		forestWind:registerObject('ForestWindEmitter_MKDW')
	end
	
	-- initialize ForestWindEmitter
	initForestWind()
	
	
	-- initialize physics wind
	initPhysicsWind()
	
	-- set correct starting ground model
	changeGroundModel(false) --> causing buggy terrain ?
	
	-- roll seed
	math.randomseed(be:getPlayerVehicle(0):getPosition().x)
	
	-- display and log message
	log("I", "", "Extension loaded")
	guihooks.trigger('toastrMsg', {type = "success", title = "Dynamic Weather", msg = "Loaded successfully :D", config = {timeOut = 5000}}) 
end


-- executed whenever new level loaded -> unload extension
local prestart = false
local function onClientPreStartMission()
	log("I", "", "Changed level. Unloading extension . . .")
	guihooks.trigger('toastrMsg', {type = "warning", title = "Dynamic Weather", msg = "Changed level - dynamic weather disabled. Please restart manually.", config = {timeOut = 5000}})
	prestart = true
	extension.unload('dynamicWeatherMK_weather')
end


-- executed when lua extension unloaded -> keybord input or level unsupported or level changed
local function onExtensionUnloaded()
	if unsupported then -- if executed because level unsupported
		log("I", "", "Failed to load extension: level not supported")
		guihooks.trigger('toastrMsg', {type = "warning", title = "Dynamic Weather", msg = "The current level isn't supported :C", config = {timeOut = 5000}})
	else
		guihooks.trigger('toastrMsg', {type = "success", title = "Dynamic Weather", msg = "Dynamic weather disabled.", config = {timeOut = 5000}})
		log("I", "", "Unloading extension . . .")
	end
end

------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------------------------


-----
-- funtions to change weather components
-----

-- increases cloud coverage
local function increaseClouds() 
	clouds.coverage = clouds.coverage + 0.01
	if clouds.exposure > 0.0 then
		clouds.exposure = clouds.exposure - 0.01
		cloudColor = cloudColor - 0.004 --> minimum 0.4 = 1 - 0.6 = 1 - 0.004 * 150
		clouds.baseColor = Point4F(cloudColor, cloudColor, cloudColor, 1)
	end
	
	-- affect sunsky and fog --
	if clouds.coverage > 0.2 and clouds.coverage < 0.6 then --> cloudy
		sunsky.skyBrightness = sunsky.skyBrightness - 2.5 -- from 600 to 500 within 0.4/0.01 = 40 steps
		sunsky.brightness = sunsky.brightness - 0.00625   -- from 1.0 to 0.75 within 40 steps
		sunsky.sunSize = sunsky.sunSize - 0.00125         -- from 0.1 to 0.05
		sunsky.flareScale = sunsky.flareScale - 0.1       -- from 5.0 to 1.0
		sunsky.sunScale = Point4F(sunsky.sunScale.x - 0.00715, sunsky.sunScale.y - 0.0039, sunsky.sunScale.z - 0.00205, 1) 
		                                                  -- from (255,211,185) to (182,171,165) --> (1,0.827,0.725) to (0.714,0.671,0.647)
		-- adjust fog color
		sunsky.fogScale = Point4F(sunsky.fogScale.x - 0.001775, sunsky.fogScale.y - 0.003325, sunsky.fogScale.z - 0.0051,1) 
														  -- from (131,187,255,255) to (113,153,203,255) --> (0.514,0.733,1,1) to (0.443,0.600,0.796,1)	
														  
	elseif clouds.coverage > 0.6 and clouds.coverage < 1.0 then --> rainy, foggy, both
		sunsky.skyBrightness = sunsky.skyBrightness - 3.75      -- from 500 to 350 within 0.4/0.01 = 40 steps 
		sunsky.brightness = sunsky.brightness - 0.00875         -- from 0.75 to 0.4 within 40 steps
		-- prevent negative values!
		if sunsky.sunSize > 0.004 then
			sunsky.sunSize = sunsky.sunSize - 0.0025            -- from 0.05 to 0.0 within 20 steps
		else
			sunsky.sunSize = 0
			--guihooks.trigger('toastrMsg', {type = "warning", title = "Dynamic Weather", msg = "Prevented negative sun size!", config = {timeOut = 5000}})
		end
		if sunsky.flareScale > 0.08 then                        -- from 1.0 to 0.0 within 20 steps
			sunsky.flareScale = sunsky.flareScale - 0.05
		else
			sunsky.flareScale = 0
			--guihooks.trigger('toastrMsg', {type = "warning", title = "Dynamic Weather", msg = "Prevented negative sunflare scale!", config = {timeOut = 5000}})
		end
		sunsky.sunScale = Point4F(sunsky.sunScale.x - 0.0108, sunsky.sunScale.y - 0.009725, sunsky.sunScale.z - 0.009125, 1) 
																-- from (182,171,165) to (72,72,72) --> (0.714,0.671,0.647) to (0.282,0.282,0.282)
		-- adjust fog color
		sunsky.fogScale = Point4F(sunsky.fogScale.x - 0.0002, sunsky.fogScale.y - 0.0008, sunsky.fogScale.z - 0.001475, 1)
																-- from (113,153,203) to (111,145,188) --> (0.443,0.600,0.796) to (0.435,0.568,0.737)
		
	elseif clouds.coverage > 1.0 then --> stormy --> dark clouds
		sunsky.skyBrightness = sunsky.skyBrightness - 5.0 -- from 350 to 100 within 0.5/0.01 = 50 steps
		sunsky.brightness = sunsky.brightness - 0.004	  -- from 0.4 to 0.2 within 50 steps
		sunsky.sunScale = Point4F(sunsky.sunScale.x - 0.00306, sunsky.sunScale.y - 0.00306, sunsky.sunScale.z - 0.00306, 1)
														  -- from (72,72,72) to (33,33,33) --> (0.282,0.282,0.282) to (0.129,0.129,0.129)
		-- adjust fog color												  
		sunsky.sunScale = Point4F(sunsky.sunScale.x - 0.00164, sunsky.sunScale.y - 0.00406, sunsky.sunScale.z - 0.00706, 1)
														  -- from (111,145,188) to (90,93,98) --> (0.435,0.568,0.737) to (0.353,0.365,0.384)
	end 
	
	--print("clouds")
	
	-- apply changes
	scenetree.findObjectById(sunsky:getId()):postApply()
	scenetree.findObjectById(clouds:getId()):postApply()
end


-- decreases cloud coverage
local function decreaseClouds()
	if clouds.coverage > 0.0 then
		clouds.coverage = clouds.coverage - 0.01
		clouds.exposure = clouds.exposure + 0.01
		cloudColor = cloudColor + 0.004 -- wait, WHAT??? --> mistake ? --> Fixed :)
		clouds.baseColor = Point4F(cloudColor, cloudColor, cloudColor, 1)
	end
	
	-- affect sunsky and fog
	if clouds.coverage > 0.2 and clouds.coverage < 0.6 then --> sunny
		sunsky.skyBrightness = sunsky.skyBrightness + 2.5 -- from 600 to 500 within 0.4/0.01 = 40 steps
		sunsky.brightness = sunsky.brightness + 0.00625   -- from 1.0 to 0.75 within 40 steps
		sunsky.sunSize = sunsky.sunSize + 0.00125         -- from 0.1 to 0.05
		sunsky.flareScale = sunsky.flareScale + 0.1       -- from 5.0 to 1.0
		sunsky.sunScale = Point4F(sunsky.sunScale.x + 0.00715, sunsky.sunScale.y + 0.0039, sunsky.sunScale.z + 0.00205, 1) 
		                                                  -- from (255,211,185) to (182,171,165) --> (1,0.827,0.725) to (0.714,0.671,0.647)
		-- adjust fog color
		sunsky.fogScale = Point4F(sunsky.fogScale.x + 0.001775, sunsky.fogScale.y + 0.003325, sunsky.fogScale.z + 0.0051,1) 
														  -- from (131,187,255,255) to (113,153,203,255) --> (0.514,0.733,1,1) to (0.443,0.600,0.796,1)												  
	
	elseif clouds.coverage > 0.6 and clouds.coverage < 1.0 then --> cloudy
		sunsky.skyBrightness = sunsky.skyBrightness + 3.75      -- from 500 to 350 within 0.4/0.01 = 40 steps 
		sunsky.brightness = sunsky.brightness + 0.00875         -- from 0.75 to 0.4 within 40 steps
		-- prevent negative values!
		if clouds.coverage < 0.79 then
			sunsky.sunSize = sunsky.sunSize + 0.0025          -- from 0.05 to 0.0 within 20 steps
		else
			sunsky.sunSize = 0
		end
		if clouds.coverage < 0.79 then                        -- from 1.0 to 0.0 within 20 steps
			sunsky.flareScale = sunsky.flareScale + 0.05
		else
			sunsky.flareScale = 0
		end
		sunsky.sunScale = Point4F(sunsky.sunScale.x + 0.0108, sunsky.sunScale.y + 0.009725, sunsky.sunScale.z + 0.009125, 1) 
																-- from (182,171,165) to (72,72,72) --> (0.714,0.671,0.647) to (0.282,0.282,0.282)
		-- adjust fog color
		sunsky.fogScale = Point4F(sunsky.fogScale.x + 0.0002, sunsky.fogScale.y + 0.0008, sunsky.fogScale.z + 0.001475, 1)
		-- from (113,153,203) to (111,145,188) --> (0.443,0.600,0.796) to (0.435,0.568,0.737)
		
	elseif clouds.coverage > 1.0 then --> rainy, foggy, both
		sunsky.skyBrightness = sunsky.skyBrightness + 5.0 -- from 350 to 100 within 0.5/0.01 = 50 steps
		sunsky.brightness = sunsky.brightness + 0.004	  -- from 0.4 to 0.2 within 50 steps
		sunsky.sunScale = Point4F(sunsky.sunScale.x + 0.00306, sunsky.sunScale.y + 0.00306, sunsky.sunScale.z + 0.00306, 1)
														  -- from (72,72,72) to (33,33,33) --> (0.282,0.282,0.282) to (0.129,0.129,0.129)
		-- adjust fog color												  
		sunsky.sunScale = Point4F(sunsky.sunScale.x + 0.00164, sunsky.sunScale.y + 0.00406, sunsky.sunScale.z + 0.00706, 1)
														  -- from (111,145,188) to (90,93,98) --> (0.435,0.568,0.737) to (0.353,0.365,0.384)
	
	end 
	
	--print("clouds")
	
	-- apply changes
	scenetree.findObjectById(sunsky:getId()):postApply()
	scenetree.findObjectById(clouds:getId()):postApply()
end


-- increases fog density --> TODO: affect windSpeed -> NO!, done within rollSlightChange()
local function increaseFog(times)
	for i = 1, times do
		fog.fogDensity = fog.fogDensity + 0.0001 --> coupled
		if fog.fogDensity > 0.01 then            --     ||   
			fog.visibleDistance = fog.visibleDistance - 4.4 --> from 5000m to about 1000m
		end
	end
	
	-- apply changes
	scenetree.findObjectById(fog:getId()):postApply()
	scenetree.findObjectById(sunsky:getId()):postApply()
end


-- decreases fog density --> TODO: affect windSpeed -> NO!, done within rollSlightChange()
local function decreaseFog(times)
	-- avoid negative or generally too low fog density
	local newFog = fog.fogDensity - 0.0001 * times
	if newFog < fogGoal[currentEvent] then newFog = fogGoal[currentEvent] goto skipLoop end
	for i = 0, times do
		fog.fogDensity = fog.fogDensity - 0.0001
		if fog.fogDensity > 0.01 then
			fog.visibleDistance = fog.visibleDistance + 4.4 --> from about 1000m to 5000m
		end
	end
	goto skipNewFog
	
	:: skipLoop ::
	fog.fogDensity = newFog
	
	:: skipNewFog ::
	
	-- apply changes
	scenetree.findObjectById(fog:getId()):postApply()
	scenetree.findObjectById(sunsky:getId()):postApply()
end


-- increases wind speed
local function increaseWind()
	if isWindBursting then return end
	-- set forest wind
	forestWind.strength = forestWind.strength + 0.01
	scenetree.findObjectById(forestWind:getId()):postApply()
	-- set physics wind
	physicsWind = Point3F(forestWind.strength * forestToPhysicsFactor, forestWind.strength * forestToPhysicsFactor, 0)
	-- set cloud speed
	clouds.windSpeed = clouds.windSpeed + 0.003
	scenetree.findObjectById(clouds:getId()):postApply()
end


-- decreases wind speed
local function decreaseWind()
	if isWindBursting then return end
	if forestWind.strength > 0.0 then
		-- set forest wind
		forestWind.strength = forestWind.strength - 0.01
		scenetree.findObjectById(forestWind:getId()):postApply()
		-- set physics wind
		physicsWind = Point3F(forestWind.strength * forestToPhysicsFactor, forestWind.strength * forestToPhysicsFactor, 0)
		-- set cloud speed
		clouds.windSpeed = clouds.windSpeed - 0.003
		scenetree.findObjectById(clouds:getId()):postApply()
	end
end


-- increases rain amount and volume
local function increaseRain()
	rain.numDrops = rain.numDrops + 20
	rainSound.volume = rain.numDrops / 2000 --> reach max volume of 1.0 with 2000 drops
	
	-- check for ground model change
	if rainBelowTreshold then
		if rain.numDrops > rainTreshold then
			rainBelowTreshold = false
			rainAboveTreshold = true
			changeGroundModel(true) --> change to wet
		end
	end
	
	-- apply changes
	scenetree.findObjectById(rain:getId()):postApply()
	scenetree.findObjectById(rainSound:getId()):postApply()
end


-- decreases rain amount and volume
local function decreaseRain()
	-- avoid negative values
	if rain.numDrops > 20 then
		rain.numDrops = rain.numDrops - 20
		rainSound.volume = rain.numDrops / 2000 --> reach min volume of 0.0 with 0 drops
		
		-- check for ground model change
		if rainAboveTreshold then
			if rain.numDrops < (rainTreshold - 200) then --> stay wet a while longer
				rainBelowTreshold = true
				rainAboveTreshold = false
				changeGroundModel(false) --> change to dry
			end
		end
	
	else
		rain.numDrops = 0
		rainSound.volume = 0.0
	end
	
	-- apply changes
	scenetree.findObjectById(rain:getId()):postApply()
	scenetree.findObjectById(rainSound:getId()):postApply()
end




-----
-- dynamic weather logic
-----
-- {"name":"theLevelInfo","class":"LevelInfo","persistentId":"aa399a62-e9f8-11e6-96ae-bcd60d6fa220","__parent":"Level_objects","canvasClearColor":"White","enabled":"1","fogAtmosphereHeight":2500,"fogColor":[0.662744999,0.886274993,0.996078014,1],"fogDensity":0.000250000012,"fogDensityOffset":8,"globalEnviromentMap":"cubemap_italy_reflection","gravity":-9.81000042,"temperatureCurveC":[0,27,0.200000003,26,0.5,23,0.800000012,24,1,27],"visibleDistance":25000}
-- {"name":"sunsky","class":"ScatterSky","persistentId":"25209345-9b9c-4a93-acda-789aec4adc2e","__parent":"Level_objects","position":[70.3971024,4.53920984,73.9923019],"ambientScale":[1,0.894117653,0.78039217,1],"ambientScaleGradientFile":"art/sky_gradients/default/gradient_ambient.png","azimuth":52.9285278,"colorize":[0.215686277,0.349019617,0.603921592,1],"colorizeGradientFile":"art/sky_gradients/default/gradient_colorize.png","elevation":55.432579,"enableFogFallBack":false,"fadeStartDistance":1000,"flareScale":5,"flareType":"BNG_Sunflare_3","fogScale":[0.396078438,0.666666687,1,1],"fogScaleGradientFile":"art/sky_gradients/default/gradient_fog.png","lastSplitTerrainOnly":true,"logWeight":0.980000019,"mieScattering":0.000634325785,"moonLightColor":[0.0980392024,0.0980392024,0.0980392024,1],"moonMat":"Moon_Glow_Mat","moonScale":0.0299999993,"nightColor":[1,0.894117653,0.78039217,1],"nightCubemap":"nightCubemap","nightFogColor":[0.396078438,0.666666687,1,1],"nightFogGradientFile":"art/sky_gradients/default/gradient_fog.png","nightGradientFile":"art/sky_gradients/default/gradient_ambient.png","occlusionScale":0.0250000004,"overDarkFactor":[40000,8000,5000,650],"shadowDarkenColor":[0,0,0,0],"shadowDistance":1600,"shadowSoftness":0.200000003,"skyBrightness":40,"sunScale":[0.996078432,0.870588243,0.784313738,1],"sunScaleGradientFile":"art/sky_gradients/default/gradient_sunscale.png","texSize":1024,"useNightCubemap":true}
-- {"name":"clouds1","class":"CloudLayer","persistentId":"0066e259-97dd-4dd8-b878-4f42ec7a42d0","__parent":"Level_objects","position":[-963.020386,1032.68262,-167.30658],"Textures":[{},{},{}],"coverage":0.100000001,"exposure":1.39999998,"height":8,"rotationMatrix":[1,0,0,0,0.999999762,-0.000690533896,0,0.000690533896,0.999999762],"texture":"levels/italy/art/skies/SkyNormals_05.dds","windSpeed":0.150000006}
-- {"class":"ForestWindEmitter","persistentId":"4eab870d-5ffc-47ad-85d8-f806e1524197","__parent":"Level_objects","rotationMatrix":[1,0,0,0,0.999999762,-0.000690533896,0,0.000690533896,0.999999762],"strength":0.5}


-- random wind bursts
local maxBurstMultiplier = 1.5 --> standard value for sunny weather, 0.2 * 2.5 * 25 = 13 mph = 20 kph kph avg, stormy: 1.0 * 4 * 25 = 100 mph = 160 kph max
local maxBurstTime = 0 --> time of burst in seconds
local maxBurstDistance = 120  --> max time in seconds between wind bursts
local minBurstDistance = 30 --> min time in seconds between wind bursts
local burstTimer = math.random(30,60) --> start value
--local burstStrength = 0 --> strength of burst in kph, not used
local cacheOldWind = 0.2

local function doWindBurst() -- TODO: make bursts more gradual ?
	local value = ( math.random() + math.random(1,maxBurstMultiplier) ) * forestWind.strength
	maxBurstTime = math.random(2,5) --> burst between 2 and 5 seconds
	burstTimer = math.random(minBurstDistance, maxBurstDistance) --> time until next burst
	
	isWindBursting = true
	cacheOldWind = forestWind.strength
	
	-- apply value
	forestWind.strength = value
	physicsWind.x = value * forestToPhysicsFactor
	physicsWind.y = value * forestToPhysicsFactor
	scenetree.findObjectById(forestWind:getId()):postApply()
	
	if doMessageDisplay then
	guihooks.trigger('Message',{msg = "Wind burst of " .. maxBurstTime .. " seconds and " .. string.format("%.2f",physicsWind.x) .. " kph!", category = "dynamicWeather", icon = "check_circle", ttl=10})
	end
end


-- slight weather change, every 15 seconds
local function rollSlightChange(value)

	local delta = math.random() --> random change, +/-
	print("rolled slight change with " .. value)
	
	-- cloud coverage
	if value == 1 then 
		if delta > 0.5 then -- more clouds
			if clouds.coverage < events[currentEvent].cloudCover then
				increaseClouds()
			end
		elseif currentEvent == "sunny" or currentEvent == "cloudy" or clouds.coverage > 0.82 then -- less clouds --> don't show sun when foggy/rainy
			decreaseClouds()
		end
	-- wind speed
	elseif value == 2 then
		if delta > 0.5 then -- stronger wind
			if forestWind.strength < events[currentEvent].windSpeed then
				increaseWind()
			end
		else -- softer wind
			decreaseWind()
		end
		--print("wind strength changed to " .. forestWind.strength)
	-- rain, if isRainEvent
	elseif value == 3 and isRainEvent then
		if delta > 0.5 then -- more rain
			if rain.numDrops < events[currentEvent].rainAmount then
				increaseRain()
			end
		else --less rain
			decreaseRain()
		end
	-- fog, if isFogEvent
	elseif value == 4 then
		if delta > 0.5 then -- more fog
			if fog.fogDensity < events[currentEvent].fogDensity then
				increaseFog(1)
			end
		else -- less fog
			decreaseFog(1)
		end
	end
	
	-- TODO: less random, dependent on circumstances --> couple fog with wind
	
end


-- do changes to reach queued event
local changeToEvent = false
local cloudsDone = false
local slowFog = false --> change fog more slowly
local slowFogCounter = 0
local fogTimes = 1    --> affects how fast fog is changing
local fogDone = false
local windDone = false
local rainDone = false

local function doEventStep()
	print("eventstep!")
	
	-- clouds
	if not cloudsDone then
	if clouds.coverage < cloudCoverGoal[currentEvent] then
		increaseClouds()
		if clouds.coverage >= cloudCoverGoal[currentEvent] then cloudsDone = true end --> goal reached
	else
		decreaseClouds()
		if clouds.coverage <= cloudCoverGoal[currentEvent] then cloudsDone = true end --> goal reached
	end
	end
	
	-- fog
	if not fogDone then   	
	if slowFog and (slowFogCounter < 3) then 
		slowFogCounter = slowFogCounter + 1 --> change fog density slower
		goto skip
	else
		slowFogCounter = 0
	end
	if fog.fogDensity < fogGoal[currentEvent] then
		increaseFog(fogTimes) --> one time
		if fog.fogDensity >= fogGoal[currentEvent] then fogDone = true end --> goal reached
	else
		decreaseFog(fogTimes) --> one time
		if fog.fogDensity <= fogGoal[currentEvent] then fogDone = true end --> goal reached
	end
	end
	
	::skip::
	
	-- wind
	if not windDone then
	if forestWind.strength < windGoal[currentEvent] then
		increaseWind()
		if forestWind.strength >= windGoal[currentEvent] then windDone = true end --> goal reached
	else
		decreaseWind()
		if forestWind.strength <= windGoal[currentEvent] then windDone = true end --> goal reached
	end
	end
	
	-- rain --> numDrops and volume
	if not rainDone then
	if rain.numDrops < rainGoal[currentEvent] then
		increaseRain()
		if rain.numDrops >= rainGoal[currentEvent] then rainDone = true end --> goal reached
	else
		decreaseRain()
		if rain.numDrops <= rainGoal[currentEvent] then rainDone = true end --> goal reached
	end
	end
	
	if cloudsDone and fogDone and windDone and rainDone then
		print("Event changed!")
		changeToEvent = false
		slowFog = false
		
		-- set variables undone
		cloudsDone = false
		fogDone = false
		windDone = false
		rainDone = false
	end
	
end


-- set weather event and change variables
local function setWeatherEvent(event)
if not changeToEvent then -- if no event queued
	if event == "sunny" then
		isRainEvent = false
		isFogEvent = false
		changeToEvent = true
		currentEvent = event
		-- wind bursts
		maxBurstMultiplier = 1.5 --> max strength of burst, depending on current windSpeed and randomness
		minBurstDistance = 30
		maxBurstDistance = 120 --> seconds inbetween bursts
	elseif event == "cloudy" then
		isRainEvent = true
		isFogEvent = false
		changeToEvent = true
		currentEvent = event
		maxBurstMultiplier = 2
		minBurstDistance = 20
		maxBurstDistance = 100
	elseif event == "lightRain" then
		isRainEvent = true
		isFogEvent = true
		changeToEvent = true
		currentEvent = event
		maxBurstMultiplier = 2.5
		minBurstDistance = 20
		maxBurstDistance = 95
	elseif event == "rainy" then
		isRainEvent = true
		isFogEvent = true
		changeToEvent = true
		currentEvent = event
		maxBurstMultiplier = 2.5
		minBurstDistance = 20
		maxBurstDistance = 90
	elseif event == "foggy" then
		isRainEvent = true
		isFogEvent = true
		changeToEvent = true
		currentEvent = event
		maxBurstMultiplier = 1
		minBurstDistance = 60
		maxBurstDistance = 180
	elseif event == "foggyRain" then
		isRainEvent = true
		isFogEvent = true
		changeToEvent = true
		currentEvent = event
		maxBurstMultiplier = 1.5
		minBurstDistance = 30
		maxBurstDistance = 150
	elseif event == "stormy" then
		--log("E", "", "Stormy weather is not implemented yet. Have I not told you so?")
		--goto stormSkip
		isRainEvent = true
		isFogEvent = true
		changeToEvent = true
		currentEvent = event
		maxBurstMultiplier = 3
		minBurstDistance = 10
		maxBurstDistance = 40
	end
	
	if changeToEvent then
		log("I", "", "Changed weather event to " .. event)
	else
		log("E", "", "Unknown event. Please check for spelling mistakes.")
		return
	end
	
	:: stormSkip ::
	
else -- if already changing to event
	log("E", "", "Another event is already queued!")
end
end


-- queue weather event to change to
local function queueEvent(value) --> value is math.random()
if not changeToEvent then --> if not already changing

	local eventMessage = "Could not change weather event"

	if currentEvent == "sunny" then --> if sunny, go cloudy or lightRain
		if value > 0.25 then
			slowFog = true --> change fog more slowly
			fogTimes = 1
			eventMessage = "Clouds are gathering."
			setWeatherEvent("cloudy")
		else
			slowFog = true --> change fog more slowly
			fogTimes = 1
			eventMessage = "Light rain is about to fall."
			setWeatherEvent("lightRain")
		end
		
	elseif currentEvent == "cloudy" then --> if cloudy, go sunny, foggy, lightRain or rainy
		if value > 0.65 then
			slowFog = true --> change fog more slowly
			fogTimes = 1
			eventMessage = "The clouds break open."
			setWeatherEvent("sunny")
		elseif value > 0.4 then
			fogTimes = 1 
			eventMessage = "The weather today: mostly cloudy, but light rain is to be expected."
			setWeatherEvent("lightRain")
		elseif value > 0.1 then
			fogTimes = 1 
			eventMessage = "The weather today: rain is to be expected."
			setWeatherEvent("rainy")
		else
			fogTimes = 2
			eventMessage = "Dense fog incoming."
			setWeatherEvent("foggy")
		end
		
	elseif currentEvent == "lightRain" then --> if lightRain, go cloudy, foggy, rainy or sunny
		if value > 0.9 then
			slowFog = true --> change fog more slowly
			fogTimes = 1
			eventMessage = "The clouds break open and rain subsides. No rainbow included :C"
			setWeatherEvent("sunny")
		elseif value > 0.5 then
			fogTimes = 1
			eventMessage = "Rain subsides within the next few minutes."
			setWeatherEvent("cloudy")
		elseif value > 0.2 then
			fogTimes = 1
			eventMessage = "Rain is about to become stronger."
			setWeatherEvent("rainy")
		else
			fogTimes = 2
			eventMessage = "Foggy news! Rainfall is about to end."
			setWeatherEvent("foggy")
		end
		
	elseif currentEvent == "foggy" then --> if foggy, go cloudy, rainy, lightRain, sunny or foggyRain
		if value > 0.85 then 
			fogTimes = 3 
			eventMessage = "From fog to rain. A british tradition!"
			setWeatherEvent("rainy")
		elseif value > 0.65 then
			fogTimes = 3 
			eventMessage = "Fog will clear and rain will come."
			setWeatherEvent("lightRain")
		elseif value > 0.5 then
			fogTimes = 1 
			eventMessage = "And then the rain began to fall :C"
			setWeatherEvent("foggyRain")
		elseif value > 0.2 then
			fogTimes = 5 
			eventMessage = "The weather today: fog is about to clear, but clouds will stay."
			setWeatherEvent("cloudy")
		else
			fogTimes = 3 
			eventMessage = "Sunshine! :D"
			setWeatherEvent("sunny")
		end
		
	elseif currentEvent == "rainy" then --> if rainy, go cloudy, lightRain, foggyRain or stormy
		if value > 0.8 then
			fogTimes = 2 
			eventMessage = "The weather today: rain subsides, but clouds are expected to stay."
			setWeatherEvent("cloudy")
		elseif value > 0.45 then
			fogTimes = 1
			eventMessage = "Rain is expected to get lighter."
			setWeatherEvent("lightRain")
		elseif value > 0.3 then
			fogTimes = 3
			eventMessage = "As if rain was not enough, now there is going to be fog as well >:C"
			setWeatherEvent("foggyRain")
		else
			fogTimes = 1
			eventMessage = "Attention! A storm is rising."
			setWeatherEvent("stormy")
		end
	
	elseif currentEvent == "foggyRain" then --> if foggyRain, go foggy, rainy or lightRain
		if value > 0.7 then
			fogTimes = 1
			eventMessage = "The weather today: rain is going to subside, but fog will stay."
			setWeatherEvent("foggy")
		elseif value > 0.3 then
			fogTimes = 3
			eventMessage = "Dense fog will clear soon."
			setWeatherEvent("lightRain")
		else
			fogTimes = 3
			eventMessage = "Frontier News: It might be raining cats and dogs soon. Please do not eat them."
			setWeatherEvent("rainy")
		end
		
	elseif currentEvent == "stormy" then --> if stormy, go rainy, lightRain, foggyRain or cloudy
		if value > 0.9 then
			fogTimes = 3
			eventMessage = "Foggy weather expected soon."
			setWeatherEvent("foggyRain")
		elseif value > 0.5 then
			fogTimes = 1
			eventMessage = "Winds are about to settle down soon."
			setWeatherEvent("rainy")
		elseif value > 0.1 then
			fogTimes = 1
			eventMessage = "Storm is about to end, but light rain will remain."
			setWeatherEvent("lightRain")
		else
			fogTimes = 2
			eventMessage = "Storm is about to end."
			setWeatherEvent("lightRain")
		end
	end
	
	
	-- TODO: change step "amount" for each event transition --> eventStepTimer --> DONE ?
	
	print("event changed to " .. currentEvent .. "!")
	if doMessageDisplay then
	guihooks.trigger('toastrMsg', {type = "success", title = "Dynamic Weather", msg = eventMessage, config = {timeOut = 5000}})
	end
	
	-- TODO: tornado event! --> value is random between 0 and 1
end
end


-- timer and randomness stuff
local timerFactor = 1.0
local applyWindTimer = 0
local slightChangeTimer = 0

--local randomFactor = be:getPlayerVehicle(0):getPosition().x
--local rerollTimer = 0 --> unused for now

local eventStepTimer = 0
local eventStepTimerTreshold = 10 --> how fast to change to event, seconds per step
--local changeToEvent = false --> defined above
local eventChance = 0
local eventChanceTimer = 0
--local eventDuration = 0 --> unused


-- every GFX frame
local function onUpdate(dt)

	timerFactor = simTimeAuthority.getReal()	--> get current simulation speed to adjust weather speed
	if simTimeAuthority.getPause() then timerFactor = 0 end
	
	
	-- increase timers
	applyWindTimer = applyWindTimer + 1 -- increased independently from simTime
	slightChangeTimer = slightChangeTimer + (dt * timerFactor)
	
	
	-- wind bursts --> TODO: make bursts more gradual ?
	if isWindBursting then 
		maxBurstTime = maxBurstTime - (dt * timerFactor)
		-- end burst
		if maxBurstTime < 0 then
			isWindBursting = false
			-- reset values
			forestWind.strength = cacheOldWind
			physicsWind.x = cacheOldWind * forestToPhysicsFactor
			physicsWind.y = cacheOldWind * forestToPhysicsFactor
		end
	else
		burstTimer = burstTimer - (dt * timerFactor)
		-- start burst
		if burstTimer <= 0 then
			doWindBurst()
		end
	end
	
	
	-- apply wind force to all physics objects each 30 frames
	if applyWindTimer > 30 then
		be:queueAllObjectLua('obj:setWind(' .. physicsWind.x .. ', ' .. physicsWind.y .. ', 0)')
		--print("Wind!" .. (forestWind.strength * forestToPhysicsFactor))
		applyWindTimer = 0
	end
	
	
	-- do slight weather change each 30 seconds, respecting slow motion --> more often ?
	if slightChangeTimer > 15 then
		if not changeToEvent then rollSlightChange(math.random(1,4)) end -- only, if no event queued
		slightChangeTimer = 0
	end
	
	
	-- weather events
	if changeToEvent then
		eventStepTimer = eventStepTimer + (dt * timerFactor)
		
		if eventStepTimer > eventStepTimerTreshold then --> should be 10 --> NO, random instead
			doEventStep()
			eventStepTimer = 0
			eventStepTimerTreshold = math.random(8,12)  --> between 8 and 12 seconds until next step
		end
	else -- elseif eventDuration <= 0 then
		eventChanceTimer = eventChanceTimer + (dt * timerFactor)
		if eventChanceTimer > 5 then --> should be 5
			
			eventChanceTimer = 0
			--print("eventChance = " .. eventChance .. " --> queues random event if greater than 100")
		
			-- increase event chance randomly each 5 seconds, respecting slow motion
			eventChance = eventChance + math.random()  -- TODO: less random ?
			
			-- queue event if eventChance high enough
			if eventChance > 60 then --> should be 60 for now
				queueEvent(math.random())
				eventChance = 0
			end
		end
	end

end




-----
-- public interface
-----

-- toggles display of messages
M.toggleMessageDisplay = function() 
							doMessageDisplay = not doMessageDisplay 
							if doMessageDisplay then
								guihooks.trigger('toastrMsg', {type = "success", title = "Dynamic Weather", msg = "Enabled messages", config = {timeOut = 5000}})
							else
								guihooks.trigger('toastrMsg', {type = "success", title = "Dynamic Weather", msg = "Disabled messages", config = {timeOut = 5000}})
							end
						 end

-- executed by game engine
M.onExtensionLoaded = onExtensionLoaded
M.onExtensionUnloaded = onExtensionUnloaded
M.onClientPreStartMission = onClientPreStartMission
M.onUpdate = onUpdate

-- for user interaction in console
M.getCurrentWeatherEvent = function() return currentEvent end
M.rollNextEvent = function() eventChance = 100 log("I", "", "Set event chance to 100, causing next event to be rolled.") end
M.setWeatherEvent = function(var) 
						if not var then log("E", "", "Please put desired weather event between the brackets") return end 
						eventChance = 0 
						setWeatherEvent(var) 
					end 
M.getEventChance = function() print(eventChance) end
										
-- for debugging purposes only, requires knowledge of underlying code
M.queueEvent = queueEvent
M.changeGroundModel = changeGroundModel 

return M